@@ -4,23 +4,37 @@ import itertools
 import logging
 import queue
 import threading
-from typing import Callable, Generic, Optional, TypeVar, Union
+from typing import Callable, Generic, Optional, Union
+
+try:
+    from typing import Literal
+except ImportError:
+    from typing_extensions import Literal  # type: ignore
 
 from aiortc import RTCPeerConnection, RTCSessionDescription
-from aiortc.contrib.media import MediaPlayer, MediaRecorder, MediaRelay
+from aiortc.contrib.media import MediaRelay
+from aiortc.mediastreams import MediaStreamTrack
 
 from .eventloop import get_server_event_loop
 from .process import (
     AsyncAudioProcessTrack,
     AsyncVideoProcessTrack,
-    AudioProcessorBase,
     AudioProcessTrack,
-    VideoProcessorBase,
     VideoProcessTrack,
-    VideoTransformerBase,
 )
 from .receive import AudioReceiver, VideoReceiver
 from .relay import get_global_relay
+from .types import (
+    AudioProcessorBase,
+    AudioProcessorFactory,
+    AudioProcessorT,
+    MediaPlayerFactory,
+    MediaRecorderFactory,
+    VideoProcessorBase,
+    VideoProcessorFactory,
+    VideoProcessorT,
+    VideoTransformerBase,
+)
 
 __all__ = [
     "AudioProcessorBase",
@@ -40,15 +54,6 @@ logger = logging.getLogger(__name__)
 logger.addHandler(logging.NullHandler())
 
 
-VideoProcessorT = TypeVar("VideoProcessorT", bound=VideoProcessorBase)
-AudioProcessorT = TypeVar("AudioProcessorT", bound=AudioProcessorBase)
-
-MediaPlayerFactory = Callable[[], MediaPlayer]
-MediaRecorderFactory = Callable[[], MediaRecorder]
-VideoProcessorFactory = Callable[[], VideoProcessorT]
-AudioProcessorFactory = Callable[[], AudioProcessorT]
-
-
 class WebRtcMode(enum.Enum):
     RECVONLY = enum.auto()
     SENDONLY = enum.auto()
@@ -59,12 +64,16 @@ class TimeoutError(Exception):
     pass
 
 
+TrackType = Literal["input:video", "input:audio", "output:video", "output:audio"]
+
+
 async def _process_offer(
     mode: WebRtcMode,
     pc: RTCPeerConnection,
     offer: RTCSessionDescription,
     relay: MediaRelay,
-    player_factory: Optional[MediaPlayerFactory],
+    source_video_track: Optional[MediaStreamTrack],
+    source_audio_track: Optional[MediaStreamTrack],
     in_recorder_factory: Optional[MediaRecorderFactory],
     out_recorder_factory: Optional[MediaRecorderFactory],
     video_processor: Optional[VideoProcessorBase],
@@ -72,13 +81,12 @@ async def _process_offer(
     video_receiver: Optional[VideoReceiver],
     audio_receiver: Optional[AudioReceiver],
     async_processing: bool,
+    sendback_video: bool,
+    sendback_audio: bool,
     callback: Callable[[Union[RTCSessionDescription, Exception]], None],
+    on_track_created: Callable[[TrackType, MediaStreamTrack], None],
 ):
     try:
-        player = None
-        if player_factory:
-            player = player_factory()
-
         in_recorder = None
         if in_recorder_factory:
             in_recorder = in_recorder_factory()
@@ -99,12 +107,19 @@ async def _process_offer(
             def on_track(input_track):
                 logger.info("Track %s received", input_track.kind)
 
+                if input_track.kind == "video":
+                    on_track_created("input:video", input_track)
+                elif input_track.kind == "audio":
+                    on_track_created("input:audio", input_track)
+
                 output_track = None
 
                 if input_track.kind == "audio":
-                    if player and player.audio:
-                        logger.info("Add player to audio track")
-                        output_track = player.audio
+                    if source_audio_track:
+                        logger.info(
+                            "Set %s as an input audio track", source_audio_track
+                        )
+                        output_track = source_audio_track
                     elif audio_processor:
                         AudioTrack = (
                             AsyncAudioProcessTrack
@@ -112,8 +127,7 @@ async def _process_offer(
                             else AudioProcessTrack
                         )
                         logger.info(
-                            "Add a input audio track %s to "
-                            "output track with audio_processor %s",
+                            "Set %s as an input audio track with audio_processor %s",
                             input_track,
                             AudioTrack,
                         )
@@ -121,13 +135,14 @@ async def _process_offer(
                             track=relay.subscribe(input_track),
                             processor=audio_processor,
                         )
-                        logger.info("Add the audio track with processor to %s", pc)
                     else:
                         output_track = input_track  # passthrough
                 elif input_track.kind == "video":
-                    if player and player.video:
-                        logger.info("Add player to video track")
-                        output_track = player.video
+                    if source_video_track:
+                        logger.info(
+                            "Set %s as an input video track", source_video_track
+                        )
+                        output_track = source_video_track
                     elif video_processor:
                         VideoTrack = (
                             AsyncVideoProcessTrack
@@ -135,8 +150,7 @@ async def _process_offer(
                             else VideoProcessTrack
                         )
                         logger.info(
-                            "Add a input video track %s to "
-                            "output track with video_processor %s",
+                            "Set %s as an input video track with video_processor %s",
                             input_track,
                             VideoTrack,
                         )
@@ -144,17 +158,24 @@ async def _process_offer(
                             track=relay.subscribe(input_track),
                             processor=video_processor,
                         )
-                        logger.info("Add the video track with processor to %s", pc)
                     else:
                         output_track = input_track
 
-                if not output_track:
-                    raise Exception(
-                        "Neither a player nor a processor is created. "
-                        "Either factory must be set."
+                if (output_track.kind == "video" and sendback_video) or (
+                    output_track.kind == "audio" and sendback_audio
+                ):
+                    logger.info(
+                        "Add a track %s of kind %s to %s",
+                        output_track,
+                        output_track.kind,
+                        pc,
+                    )
+                    pc.addTrack(relay.subscribe(output_track))
+                else:
+                    logger.info(
+                        "Block a track %s of kind %s", output_track, output_track.kind
                     )
 
-                pc.addTrack(relay.subscribe(output_track))
                 if out_recorder:
                     logger.info("Track %s is added to out_recorder", output_track.kind)
                     out_recorder.addTrack(relay.subscribe(output_track))
@@ -162,6 +183,11 @@ async def _process_offer(
                     logger.info("Track %s is added to in_recorder", input_track.kind)
                     in_recorder.addTrack(relay.subscribe(input_track))
 
+                if output_track.kind == "video":
+                    on_track_created("output:video", output_track)
+                elif output_track.kind == "audio":
+                    on_track_created("output:audio", output_track)
+
                 @input_track.on("ended")
                 async def on_ended():
                     logger.info("Track %s ended", input_track.kind)
@@ -176,6 +202,11 @@ async def _process_offer(
             def on_track(input_track):
                 logger.info("Track %s received", input_track.kind)
 
+                if input_track.kind == "video":
+                    on_track_created("input:video", input_track)
+                elif input_track.kind == "audio":
+                    on_track_created("input:audio", input_track)
+
                 if input_track.kind == "audio":
                     if audio_receiver:
                         logger.info(
@@ -208,7 +239,7 @@ async def _process_offer(
             for t in pc.getTransceivers():
                 output_track = None
                 if t.kind == "audio":
-                    if player and player.audio:
+                    if source_audio_track:
                         if audio_processor:
                             AudioTrack = (
                                 AsyncAudioProcessTrack
@@ -216,29 +247,18 @@ async def _process_offer(
                                 else AudioProcessTrack
                             )
                             logger.info(
-                                "Add a input audio track from a player %s to "
-                                "output track with audio_processor %s",
-                                player.audio,
+                                "Set %s as an input audio track "
+                                "with audio_processor %s",
+                                source_audio_track,
                                 AudioTrack,
                             )
                             output_track = AudioTrack(
-                                track=player.audio, processor=audio_processor
+                                track=source_audio_track, processor=audio_processor
                             )
-                            logger.info("Add the audio track with processor to %s", pc)
-
-                            @output_track.on("ended")
-                            async def on_ended():
-                                logger.info(
-                                    "Track %s ended. Stop its input track %s",
-                                    output_track.kind,
-                                    player.audio,
-                                )
-                                player.audio.stop()
-
                         else:
-                            output_track = player.audio  # passthrough
+                            output_track = source_audio_track  # passthrough
                 elif t.kind == "video":
-                    if player and player.video:
+                    if source_video_track:
                         if video_processor:
                             VideoTrack = (
                                 AsyncVideoProcessTrack
@@ -246,34 +266,29 @@ async def _process_offer(
                                 else VideoProcessTrack
                             )
                             logger.info(
-                                "Add a input video track from a player %s to "
-                                "output track with video_processor %s",
-                                player.video,
+                                "Set %s as an input video track "
+                                "with video_processor %s",
+                                source_video_track,
                                 VideoTrack,
                             )
                             output_track = VideoTrack(
-                                track=player.video, processor=video_processor
+                                track=source_video_track, processor=video_processor
                             )
-                            logger.info("Add the video track with processor to %s", pc)
-
-                            @output_track.on("ended")
-                            async def on_ended():
-                                logger.info(
-                                    "Track %s ended. Stop its input track %s",
-                                    output_track.kind,
-                                    player.video,
-                                )
-                                player.video.stop()
-
                         else:
-                            output_track = player.video  # passthrough
+                            output_track = source_video_track  # passthrough
 
                 if output_track:
-                    pc.addTrack(output_track)
+                    logger.info("Add a track %s to %s", output_track, pc)
+                    pc.addTrack(relay.subscribe(output_track))
                     # NOTE: Recording is not supported in this mode
                     # because connecting player to recorder does not work somehow;
                     # it generates unplayable movie files.
 
+                    if output_track.kind == "video":
+                        on_track_created("output:video", output_track)
+                    elif output_track.kind == "audio":
+                        on_track_created("output:audio", output_track)
+
         if video_receiver and video_receiver.hasTrack():
             video_receiver.start()
         if audio_receiver and audio_receiver.hasTrack():
@@ -305,6 +320,10 @@ class WebRtcWorker(Generic[VideoProcessorT, AudioProcessorT]):
     _audio_processor: Optional[AudioProcessorT]
     _video_receiver: Optional[VideoReceiver]
     _audio_receiver: Optional[AudioReceiver]
+    _input_video_track: Optional[MediaStreamTrack]
+    _input_audio_track: Optional[MediaStreamTrack]
+    _output_video_track: Optional[MediaStreamTrack]
+    _output_audio_track: Optional[MediaStreamTrack]
 
     @property
     def video_processor(self) -> Optional[VideoProcessorT]:
@@ -322,9 +341,27 @@ class WebRtcWorker(Generic[VideoProcessorT, AudioProcessorT]):
     def audio_receiver(self) -> Optional[AudioReceiver]:
         return self._audio_receiver
 
+    @property
+    def input_video_track(self) -> Optional[MediaStreamTrack]:
+        return self._input_video_track
+
+    @property
+    def input_audio_track(self) -> Optional[MediaStreamTrack]:
+        return self._input_audio_track
+
+    @property
+    def output_video_track(self) -> Optional[MediaStreamTrack]:
+        return self._output_video_track
+
+    @property
+    def output_audio_track(self) -> Optional[MediaStreamTrack]:
+        return self._output_audio_track
+
     def __init__(
         self,
         mode: WebRtcMode,
+        source_video_track: Optional[MediaStreamTrack] = None,
+        source_audio_track: Optional[MediaStreamTrack] = None,
         player_factory: Optional[MediaPlayerFactory] = None,
         in_recorder_factory: Optional[MediaRecorderFactory] = None,
         out_recorder_factory: Optional[MediaRecorderFactory] = None,
@@ -337,12 +374,16 @@ class WebRtcWorker(Generic[VideoProcessorT, AudioProcessorT]):
         async_processing: bool = True,
         video_receiver_size: int = 4,
         audio_receiver_size: int = 4,
+        sendback_video: bool = True,
+        sendback_audio: bool = True,
     ) -> None:
         self._webrtc_thread = None
         self.pc = RTCPeerConnection()
         self._answer_queue = queue.Queue()
 
         self.mode = mode
+        self.source_video_track = source_video_track
+        self.source_audio_track = source_audio_track
         self.player_factory = player_factory
         self.in_recorder_factory = in_recorder_factory
         self.out_recorder_factory = out_recorder_factory
@@ -351,11 +392,17 @@ class WebRtcWorker(Generic[VideoProcessorT, AudioProcessorT]):
         self.async_processing = async_processing
         self.video_receiver_size = video_receiver_size
         self.audio_receiver_size = audio_receiver_size
+        self.sendback_video = sendback_video
+        self.sendback_audio = sendback_audio
 
         self._video_processor = None
         self._audio_processor = None
         self._video_receiver = None
         self._audio_receiver = None
+        self._input_video_track = None
+        self._input_audio_track = None
+        self._output_video_track = None
+        self._output_audio_track = None
 
     def _run_webrtc_thread(
         self,
@@ -381,12 +428,23 @@ class WebRtcWorker(Generic[VideoProcessorT, AudioProcessorT]):
         )
 
         loop = get_server_event_loop()
+        asyncio.set_event_loop(loop)
 
         offer = RTCSessionDescription(sdp, type_)
 
         def callback(localDescription):
             self._answer_queue.put(localDescription)
 
+        def on_track_created(track_type: TrackType, track: MediaStreamTrack):
+            if track_type == "input:video":
+                self._input_video_track = track
+            elif track_type == "input:audio":
+                self._input_audio_track = track
+            elif track_type == "output:video":
+                self._output_video_track = track
+            elif track_type == "output:audio":
+                self._output_audio_track = track
+
         video_processor = None
         if self.video_processor_factory:
             video_processor = self.video_processor_factory()
@@ -408,6 +466,20 @@ class WebRtcWorker(Generic[VideoProcessorT, AudioProcessorT]):
 
         relay = get_global_relay()
 
+        source_audio_track = None
+        source_video_track = None
+        if self.player_factory:
+            player = self.player_factory()
+            if player.audio:
+                source_audio_track = relay.subscribe(player.audio)
+            if player.video:
+                source_video_track = relay.subscribe(player.video)
+        else:
+            if self.source_video_track:
+                source_video_track = relay.subscribe(self.source_video_track)
+            if self.source_audio_track:
+                source_audio_track = relay.subscribe(self.source_audio_track)
+
         @self.pc.on("iceconnectionstatechange")
         async def on_iceconnectionstatechange():
             iceConnectionState = self.pc.iceConnectionState
@@ -420,7 +492,8 @@ class WebRtcWorker(Generic[VideoProcessorT, AudioProcessorT]):
                 self.pc,
                 offer,
                 relay=relay,
-                player_factory=self.player_factory,
+                source_video_track=source_video_track,
+                source_audio_track=source_audio_track,
                 in_recorder_factory=self.in_recorder_factory,
                 out_recorder_factory=self.out_recorder_factory,
                 video_processor=video_processor,
@@ -428,7 +501,10 @@ class WebRtcWorker(Generic[VideoProcessorT, AudioProcessorT]):
                 video_receiver=video_receiver,
                 audio_receiver=audio_receiver,
                 async_processing=self.async_processing,
+                sendback_video=self.sendback_video,
+                sendback_audio=self.sendback_audio,
                 callback=callback,
+                on_track_created=on_track_created,
             )
         )
 
